from dataclasses import dataclass

import numpy as np
from numpy.typing import NDArray

from optimagic.parameters.scaling import ScalingOptions
from optimagic.parameters.space_conversion import InternalParams


@dataclass(frozen=True)
class ScaleConverter:
    factor: NDArray[np.float64] | None
    offset: NDArray[np.float64] | None

    def params_to_internal(self, vec: NDArray[np.float64]) -> NDArray[np.float64]:
        """Scale a parameter vector from external scale to internal one."""
        if self.offset is not None:
            vec = vec - self.offset
        if self.factor is not None:
            vec = vec / self.factor
        return vec

    def params_from_internal(self, vec: NDArray[np.float64]) -> NDArray[np.float64]:
        """Scale a parameter vector from internal scale to external one."""
        if self.factor is not None:
            vec = vec * self.factor
        if self.offset is not None:
            vec = vec + self.offset
        return vec

    def derivative_to_internal(
        self, derivative: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Scale a derivative vector from external scale to internal one."""
        if self.factor is not None:
            derivative = derivative * self.factor
        return derivative

    def derivative_from_internal(
        self, derivative: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Scale a derivative vector from internal scale to external one."""
        if self.factor is not None:
            derivative = derivative / self.factor
        return derivative


def get_scale_converter(
    internal_params: InternalParams,
    scaling: ScalingOptions | None,
) -> tuple[ScaleConverter, InternalParams]:
    """Get a converter between scaled and unscaled parameters.

    Args:
        internal_params: NamedTuple of internal and possibly reparametrized but not yet
            scaled parameter values and bounds.
        scaling: Scaling options. If None, no scaling is performed.

    Returns:
        ScaleConverter: Dataclass with methods to convert between scaled and unscaled
            internal parameters and derivatives.
        InternalParams: Dataclass with internal parameter values and bounds.

    """
    # fast path
    if scaling is None:
        return ScaleConverter(factor=None, offset=None), internal_params

    factor, offset = calculate_scaling_factor_and_offset(
        internal_params=internal_params,
        options=scaling,
    )

    converter = ScaleConverter(factor=factor, offset=offset)

    if internal_params.soft_lower_bounds is not None:
        _soft_lower = converter.params_to_internal(internal_params.soft_lower_bounds)
    else:
        _soft_lower = None

    if internal_params.soft_upper_bounds is not None:
        _soft_upper = converter.params_to_internal(internal_params.soft_upper_bounds)
    else:
        _soft_upper = None

    params = InternalParams(
        values=converter.params_to_internal(internal_params.values),
        lower_bounds=converter.params_to_internal(internal_params.lower_bounds),
        upper_bounds=converter.params_to_internal(internal_params.upper_bounds),
        names=internal_params.names,
        soft_lower_bounds=_soft_lower,
        soft_upper_bounds=_soft_upper,
    )

    return converter, params


def calculate_scaling_factor_and_offset(
    internal_params: InternalParams,
    options: ScalingOptions,
) -> tuple[NDArray[np.float64], NDArray[np.float64] | None]:
    x = internal_params.values
    lower_bounds = internal_params.lower_bounds
    upper_bounds = internal_params.upper_bounds

    if options.method == "start_values":
        raw_factor = np.clip(np.abs(x), options.clipping_value, np.inf)
        scaling_offset = None
    elif options.method == "bounds":
        raw_factor = upper_bounds - lower_bounds
        scaling_offset = lower_bounds
    else:
        raise ValueError(f"Invalid scaling method: {options.method}")

    scaling_factor = raw_factor / options.magnitude

    return scaling_factor, scaling_offset



"""Public functions for optimization.

This module defines the public functions `maximize` and `minimize` that will be called
by users.

Internally, `maximize` and `minimize` just call `create_optimization_problem` with
all arguments and add the `direction`. In `create_optimization_problem`, the user input
is consolidated and converted to stricter types.  The resulting `OptimizationProblem`
is then passed to `_optimize` which handles the optimization logic.

`_optimize` processes the optimization problem and performs the actual optimization.

"""

from __future__ import annotations

from pathlib import Path
from typing import Any, Callable, Sequence, Type, cast

from scipy.optimize import Bounds as ScipyBounds

from optimagic.batch_evaluators import process_batch_evaluator
from optimagic.constraints import Constraint
from optimagic.differentiation.numdiff_options import NumdiffOptions, NumdiffOptionsDict
from optimagic.exceptions import (
    InvalidFunctionError,
)
from optimagic.logging.logger import LogReader, LogStore
from optimagic.logging.types import ProblemInitialization
from optimagic.optimization.algorithm import Algorithm
from optimagic.optimization.create_optimization_problem import (
    OptimizationProblem,
    create_optimization_problem,
)
from optimagic.optimization.error_penalty import get_error_penalty_function
from optimagic.optimization.fun_value import FunctionValue
from optimagic.optimization.internal_optimization_problem import (
    InternalBounds,
    InternalOptimizationProblem,
)
from optimagic.optimization.multistart import (
    run_multistart_optimization,
)
from optimagic.optimization.multistart_options import (
    MultistartOptions,
    MultistartOptionsDict,
    get_internal_multistart_options_from_public,
)
from optimagic.optimization.optimization_logging import log_scheduled_steps_and_get_ids
from optimagic.optimization.optimize_result import OptimizeResult
from optimagic.optimization.process_results import (
    ExtraResultFields,
    process_multistart_result,
    process_single_result,
)
from optimagic.parameters.bounds import Bounds
from optimagic.parameters.conversion import (
    get_converter,
)
from optimagic.parameters.nonlinear_constraints import process_nonlinear_constraints
from optimagic.parameters.scaling import ScalingOptions, ScalingOptionsDict
from optimagic.typing import (
    AggregationLevel,
    Direction,
    ErrorHandling,
    ErrorHandlingLiteral,
    NonNegativeFloat,
    PyTree,
)

FunType = Callable[..., float | PyTree | FunctionValue]
AlgorithmType = str | Algorithm | Type[Algorithm]
ConstraintsType = Constraint | list[Constraint] | dict[str, Any] | list[dict[str, Any]]
JacType = Callable[..., PyTree]
FunAndJacType = Callable[..., tuple[float | PyTree | FunctionValue, PyTree]]
HessType = Callable[..., PyTree]
# TODO: refine this type
CallbackType = Callable[..., Any]

CriterionType = Callable[..., float | dict[str, Any]]
CriterionAndDerivativeType = Callable[..., tuple[float | dict[str, Any], PyTree]]


from optimagic.logging.logger import LogOptions


def maximize(
    fun: FunType | CriterionType | None = None,
    params: PyTree | None = None,
    algorithm: AlgorithmType | None = None,
    *,
    bounds: Bounds | ScipyBounds | Sequence[tuple[float, float]] | None = None,
    constraints: ConstraintsType | None = None,
    fun_kwargs: dict[str, Any] | None = None,
    algo_options: dict[str, Any] | None = None,
    jac: JacType | list[JacType] | None = None,
    jac_kwargs: dict[str, Any] | None = None,
    fun_and_jac: FunAndJacType | CriterionAndDerivativeType | None = None,
    fun_and_jac_kwargs: dict[str, Any] | None = None,
    numdiff_options: NumdiffOptions | NumdiffOptionsDict | None = None,
    # TODO: add typed-dict support?
    logging: bool | str | Path | LogOptions | dict[str, Any] | None = None,
    error_handling: ErrorHandling | ErrorHandlingLiteral = ErrorHandling.RAISE,
    error_penalty: dict[str, float] | None = None,
    scaling: bool | ScalingOptions | ScalingOptionsDict = False,
    multistart: bool | MultistartOptions | MultistartOptionsDict = False,
    collect_history: bool = True,
    skip_checks: bool = False,
    # scipy aliases
    x0: PyTree | None = None,
    method: str | None = None,
    args: tuple[Any] | None = None,
    # scipy arguments that are not yet supported
    hess: HessType | None = None,
    hessp: HessType | None = None,
    callback: CallbackType | None = None,
    # scipy arguments that will never be supported
    options: dict[str, Any] | None = None,
    tol: NonNegativeFloat | None = None,
    # deprecated arguments
    criterion: CriterionType | None = None,
    criterion_kwargs: dict[str, Any] | None = None,
    derivative: JacType | None = None,
    derivative_kwargs: dict[str, Any] | None = None,
    criterion_and_derivative: CriterionAndDerivativeType | None = None,
    criterion_and_derivative_kwargs: dict[str, Any] | None = None,
    log_options: dict[str, Any] | None = None,
    lower_bounds: PyTree | None = None,
    upper_bounds: PyTree | None = None,
    soft_lower_bounds: PyTree | None = None,
    soft_upper_bounds: PyTree | None = None,
    scaling_options: dict[str, Any] | None = None,
    multistart_options: dict[str, Any] | None = None,
) -> OptimizeResult:
    """Maximize fun using algorithm subject to constraints.

    Args:
        fun: The objective function of a scalar, least-squares or likelihood
            optimization problem. Non-scalar objective functions have to be marked
            with the `mark.likelihood` or `mark.least_squares` decorators. `fun` maps
            params and fun_kwargs to an objective value. See :ref:`how-to-fun` for
            details and examples.
        params: The start parameters for the optimization. Params can be numpy arrays,
            dictionaries, pandas.Series, pandas.DataFrames, NamedTuples, floats, lists,
            and any nested combination thereof. See :ref:`params` for details and
            examples.
        algorithm: The optimization algorithm to use. Can be a string, subclass of
            :class:`optimagic.Algorithm` or an instance of a subclass of
            :class:`optimagic.Algorithm`. For guidelines on how to choose an algorithm
            see :ref:`how-to-select-algorithms`. For examples of specifying and
            configuring algorithms see :ref:`specify-algorithm`.
        bounds: Lower and upper bounds on the parameters. The most general and preferred
            way to specify bounds is an :class:`optimagic.Bounds` object that collects
            lower, upper, soft_lower and soft_upper bounds. The soft bounds are used for
            sampling based optimizers but are not enforced during optimization. Each
            bound type mirrors the structure of params. See :ref:`how-to-bounds` for
            details and examples. If params is a flat numpy array, you can also provide
            bounds via any format that is supported by scipy.optimize.minimize.
        constraints: Constraints for the optimization problem. Constraints can be
            specified as a single :class:`optimagic.Constraint` object, a list of
            Constraint objects. For details and examples check :ref:`constraints`.
        fun_kwargs: Additional keyword arguments for the objective function.
        algo_options: Additional options for the optimization algorithm. `algo_options`
            is an alternative to configuring algorithm objects directly. See
            :ref:`list_of_algorithms` for supported options of each algorithm.
        jac: The first derivative of `fun`. Providing a closed form derivative can be
            a great way to speed up your optimization. The easiest way to get
            a derivative for your objective function are autodiff frameworks like
            JAX. For details and examples see :ref:`how-to-jac`.
        jac_kwargs: Additional keyword arguments for `jac`.
        fun_and_jac: A function that returns both the objective value and the
            derivative. This can be used do exploit synergies in the calculation of the
            function value and its derivative. For details and examples see
            :ref:`how-to-jac`.
        fun_and_jac_kwargs: Additional keyword arguments for `fun_and_jac`.
        numdiff_options: Options for numerical differentiation. Can be a dictionary
            or an instance of :class:`optimagic.NumdiffOptions`.
        logging: If None, no logging is used. If a str or pathlib.Path is provided,
            it is interpreted as path to an sqlite3 file (which typically has
            the file extension ``.db``. If the file does not exist, it will be created.
            and the optimization history will be stored in that database. For more
            customization, provide LogOptions. For details and examples see
            :ref:`how-to-logging`.
        error_handling: If "raise" or ErrorHandling.RAISE, exceptions that occur during
            the optimization are raised and the optimization is stopped. If "continue"
            or ErrorHandling.CONTINUE, exceptions are caught and the function value and
            its derivative are replaced by penalty values. The penalty values are
            constructed such that the optimizer is guided back towards the start
            parameters until a feasible region is reached and then continues the
            optimization from there. For details see  :ref:`how-to-errors`.
        error_penalty: A dictionary with the keys "slope" and "constant" that
            influences the magnitude of the penalty values. For maximization problems
            both should be negative. For details see :ref:`how-to-errors`.
        scaling: If None or False, the parameter space is not rescaled. If True,
            a heuristic is used to improve the conditioning of the optimization problem.
            To choose which heuristic is used and to customize the scaling, provide
            a dictionary or an instance of :class:`optimagic.ScalingOptions`.
            For details and examples see :ref:`scaling`.
        multistart: If None or False, no multistart approach is used. If True, the
            optimization is restarted from multiple starting points. Note that this
            requires finite bounds or soft bounds for all parameters. To customize the
            multistart approach, provide a dictionary or an instance of
            :class:`optimagic.MultistartOptions`. For details and examples see
            :ref:`how-to-multistart`.
        collect_history: If True, the optimization history is collected and returned
            in the OptimizeResult. This is required to create `criterion_plot` or
            `params_plot` from an OptimizeResult.
        skip_checks: If True, some checks are skipped to speed up the optimization.
            This is only relevant if your objective function is very fast, i.e. runs in
            a few microseconds.
        x0: Alias for params for scipy compatibility.
        method: Alternative to algorithm for scipy compatibility. With `method` you can
            select scipy optimizers via their original scipy name.
        args: Alternative to fun_kwargs for scipy compatibility.
        hess: Not yet supported.
        hessp: Not yet supported.
        callback: Not yet supported.
        options: Not yet supported.
        tol: Not yet supported.
        criterion: Deprecated. Use fun instead.
        criterion_kwargs: Deprecated. Use fun_kwargs instead.
        derivative: Deprecated. Use jac instead.
        derivative_kwargs: Deprecated. Use jac_kwargs instead.
        criterion_and_derivative: Deprecated. Use fun_and_jac instead.
        criterion_and_derivative_kwargs: Deprecated. Use fun_and_jac_kwargs instead.
        lower_bounds: Deprecated. Use bounds instead.
        upper_bounds: Deprecated. Use bounds instead.
        soft_lower_bounds: Deprecated. Use bounds instead.
        soft_upper_bounds: Deprecated. Use bounds instead.
        scaling_options: Deprecated. Use scaling instead.
        multistart_options: Deprecated. Use multistart instead.

    """
    problem = create_optimization_problem(
        direction=Direction.MAXIMIZE,
        fun=fun,
        params=params,
        bounds=bounds,
        algorithm=algorithm,
        fun_kwargs=fun_kwargs,
        constraints=constraints,
        algo_options=algo_options,
        jac=jac,
        jac_kwargs=jac_kwargs,
        fun_and_jac=fun_and_jac,
        fun_and_jac_kwargs=fun_and_jac_kwargs,
        numdiff_options=numdiff_options,
        logging=logging,
        log_options=log_options,
        error_handling=error_handling,
        error_penalty=error_penalty,
        scaling=scaling,
        multistart=multistart,
        collect_history=collect_history,
        skip_checks=skip_checks,
        # scipy aliases
        x0=x0,
        method=method,
        args=args,
        # scipy arguments that are not yet supported
        hess=hess,
        hessp=hessp,
        callback=callback,
        # scipy arguments that will never be supported
        options=options,
        tol=tol,
        # deprecated arguments
        criterion=criterion,
        criterion_kwargs=criterion_kwargs,
        derivative=derivative,
        derivative_kwargs=derivative_kwargs,
        criterion_and_derivative=criterion_and_derivative,
        criterion_and_derivative_kwargs=criterion_and_derivative_kwargs,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        soft_lower_bounds=soft_lower_bounds,
        soft_upper_bounds=soft_upper_bounds,
        scaling_options=scaling_options,
        multistart_options=multistart_options,
    )
    return _optimize(problem)


def minimize(
    fun: FunType | CriterionType | None = None,
    params: PyTree | None = None,
    algorithm: AlgorithmType | None = None,
    *,
    bounds: Bounds | ScipyBounds | Sequence[tuple[float, float]] | None = None,
    constraints: ConstraintsType | None = None,
    fun_kwargs: dict[str, Any] | None = None,
    algo_options: dict[str, Any] | None = None,
    jac: JacType | list[JacType] | None = None,
    jac_kwargs: dict[str, Any] | None = None,
    fun_and_jac: FunAndJacType | CriterionAndDerivativeType | None = None,
    fun_and_jac_kwargs: dict[str, Any] | None = None,
    numdiff_options: NumdiffOptions | NumdiffOptionsDict | None = None,
    # TODO: add typed-dict support?
    logging: bool | str | Path | LogOptions | dict[str, Any] | None = None,
    error_handling: ErrorHandling | ErrorHandlingLiteral = ErrorHandling.RAISE,
    error_penalty: dict[str, float] | None = None,
    scaling: bool | ScalingOptions | ScalingOptionsDict = False,
    multistart: bool | MultistartOptions | MultistartOptionsDict = False,
    collect_history: bool = True,
    skip_checks: bool = False,
    # scipy aliases
    x0: PyTree | None = None,
    method: str | None = None,
    args: tuple[Any] | None = None,
    # scipy arguments that are not yet supported
    hess: HessType | None = None,
    hessp: HessType | None = None,
    callback: CallbackType | None = None,
    # scipy arguments that will never be supported
    options: dict[str, Any] | None = None,
    tol: NonNegativeFloat | None = None,
    # deprecated arguments
    criterion: CriterionType | None = None,
    criterion_kwargs: dict[str, Any] | None = None,
    derivative: JacType | None = None,
    derivative_kwargs: dict[str, Any] | None = None,
    criterion_and_derivative: CriterionAndDerivativeType | None = None,
    criterion_and_derivative_kwargs: dict[str, Any] | None = None,
    log_options: dict[str, Any] | None = None,
    lower_bounds: PyTree | None = None,
    upper_bounds: PyTree | None = None,
    soft_lower_bounds: PyTree | None = None,
    soft_upper_bounds: PyTree | None = None,
    scaling_options: dict[str, Any] | None = None,
    multistart_options: dict[str, Any] | None = None,
) -> OptimizeResult:
    """Minimize criterion using algorithm subject to constraints.

    Args:
        fun: The objective function of a scalar, least-squares or likelihood
            optimization problem. Non-scalar objective functions have to be marked
            with the `mark.likelihood` or `mark.least_squares` decorators. `fun` maps
            params and fun_kwargs to an objective value. See :ref:`how-to-fun` for
            details and examples.
        params: The start parameters for the optimization. Params can be numpy arrays,
            dictionaries, pandas.Series, pandas.DataFrames, NamedTuples, floats, lists,
            and any nested combination thereof. See :ref:`params` for details and
            examples.
        algorithm: The optimization algorithm to use. Can be a string, subclass of
            :class:`optimagic.Algorithm` or an instance of a subclass of
            :class:`optimagic.Algorithm`. For guidelines on how to choose an algorithm
            see :ref:`how-to-select-algorithms`. For examples of specifying and
            configuring algorithms see :ref:`specify-algorithm`.
        bounds: Lower and upper bounds on the parameters. The most general and preferred
            way to specify bounds is an :class:`optimagic.Bounds` object that collects
            lower, upper, soft_lower and soft_upper bounds. The soft bounds are used for
            sampling based optimizers but are not enforced during optimization. Each
            bound type mirrors the structure of params. See :ref:`how-to-bounds` for
            details and examples. If params is a flat numpy array, you can also provide
            bounds via any format that is supported by scipy.optimize.minimize.
        constraints: Constraints for the optimization problem. Constraints can be
            specified as a single :class:`optimagic.Constraint` object, a list of
            Constraint objects. For details and examples check :ref:`constraints`.
        fun_kwargs: Additional keyword arguments for the objective function.
        algo_options: Additional options for the optimization algorithm. `algo_options`
            is an alternative to configuring algorithm objects directly. See
            :ref:`list_of_algorithms` for supported options of each algorithm.
        jac: The first derivative of `fun`. Providing a closed form derivative can be
            a great way to speed up your optimization. The easiest way to get
            a derivative for your objective function are autodiff frameworks like
            JAX. For details and examples see :ref:`how-to-jac`.
        jac_kwargs: Additional keyword arguments for `jac`.
        fun_and_jac: A function that returns both the objective value and the
            derivative. This can be used do exploit synergies in the calculation of the
            function value and its derivative. For details and examples see
            :ref:`how-to-jac`.
        fun_and_jac_kwargs: Additional keyword arguments for `fun_and_jac`.
        numdiff_options: Options for numerical differentiation. Can be a dictionary
            or an instance of :class:`optimagic.NumdiffOptions`.
        logging: If None, no logging is used. If a str or pathlib.Path is provided,
            it is interpreted as path to an sqlite3 file (which typically has
            the file extension ``.db``. If the file does not exist, it will be created.
            and the optimization history will be stored in that database. For more
            customization, provide LogOptions. For details and examples see
            :ref:`how-to-logging`.
        error_handling: If "raise" or ErrorHandling.RAISE, exceptions that occur during
            the optimization are raised and the optimization is stopped. If "continue"
            or ErrorHandling.CONTINUE, exceptions are caught and the function value and
            its derivative are replaced by penalty values. The penalty values are
            constructed such that the optimizer is guided back towards the start
            parameters until a feasible region is reached and then continues the
            optimization from there. For details see  :ref:`how-to-errors`.
        error_penalty: A dictionary with the keys "slope" and "constant" that
            influences the magnitude of the penalty values. For minimization problems
            both should be positive. For details see :ref:`how-to-errors`.
        scaling: If None or False, the parameter space is not rescaled. If True,
            a heuristic is used to improve the conditioning of the optimization problem.
            To choose which heuristic is used and to customize the scaling, provide
            a dictionary or an instance of :class:`optimagic.ScalingOptions`.
            For details and examples see :ref:`scaling`.
        multistart: If None or False, no multistart approach is used. If True, the
            optimization is restarted from multiple starting points. Note that this
            requires finite bounds or soft bounds for all parameters. To customize the
            multistart approach, provide a dictionary or an instance of
            :class:`optimagic.MultistartOptions`. For details and examples see
            :ref:`how-to-multistart`.
        collect_history: If True, the optimization history is collected and returned
            in the OptimizeResult. This is required to create `criterion_plot` or
            `params_plot` from an OptimizeResult.
        skip_checks: If True, some checks are skipped to speed up the optimization.
            This is only relevant if your objective function is very fast, i.e. runs in
            a few microseconds.
        x0: Alias for params for scipy compatibility.
        method: Alternative to algorithm for scipy compatibility. With `method` you can
            select scipy optimizers via their original scipy name.
        args: Alternative to fun_kwargs for scipy compatibility.
        hess: Not yet supported.
        hessp: Not yet supported.
        callback: Not yet supported.
        options: Not yet supported.
        tol: Not yet supported.
        criterion: Deprecated. Use fun instead.
        criterion_kwargs: Deprecated. Use fun_kwargs instead.
        derivative: Deprecated. Use jac instead.
        derivative_kwargs: Deprecated. Use jac_kwargs instead.
        criterion_and_derivative: Deprecated. Use fun_and_jac instead.
        criterion_and_derivative_kwargs: Deprecated. Use fun_and_jac_kwargs instead.
        lower_bounds: Deprecated. Use bounds instead.
        upper_bounds: Deprecated. Use bounds instead.
        soft_lower_bounds: Deprecated. Use bounds instead.
        soft_upper_bounds: Deprecated. Use bounds instead.
        scaling_options: Deprecated. Use scaling instead.
        multistart_options: Deprecated. Use multistart instead.

    """

    problem = create_optimization_problem(
        direction=Direction.MINIMIZE,
        fun=fun,
        params=params,
        algorithm=algorithm,
        bounds=bounds,
        fun_kwargs=fun_kwargs,
        constraints=constraints,
        algo_options=algo_options,
        jac=jac,
        jac_kwargs=jac_kwargs,
        fun_and_jac=fun_and_jac,
        fun_and_jac_kwargs=fun_and_jac_kwargs,
        numdiff_options=numdiff_options,
        logging=logging,
        error_handling=error_handling,
        error_penalty=error_penalty,
        scaling=scaling,
        multistart=multistart,
        collect_history=collect_history,
        skip_checks=skip_checks,
        # scipy aliases
        x0=x0,
        method=method,
        args=args,
        # scipy arguments that are not yet supported
        hess=hess,
        hessp=hessp,
        callback=callback,
        # scipy arguments that will never be supported
        options=options,
        tol=tol,
        # deprecated arguments
        criterion=criterion,
        criterion_kwargs=criterion_kwargs,
        derivative=derivative,
        derivative_kwargs=derivative_kwargs,
        criterion_and_derivative=criterion_and_derivative,
        criterion_and_derivative_kwargs=criterion_and_derivative_kwargs,
        lower_bounds=lower_bounds,
        log_options=log_options,
        upper_bounds=upper_bounds,
        soft_lower_bounds=soft_lower_bounds,
        soft_upper_bounds=soft_upper_bounds,
        scaling_options=scaling_options,
        multistart_options=multistart_options,
    )
    return _optimize(problem)


def _optimize(problem: OptimizationProblem) -> OptimizeResult:
    """Solve an optimization problem."""
    # ==================================================================================
    # Split constraints into nonlinear and reparametrization parts
    # ==================================================================================
    constraints = problem.constraints

    nonlinear_constraints = [c for c in constraints if c["type"] == "nonlinear"]

    if nonlinear_constraints:
        if not problem.algorithm.algo_info.supports_nonlinear_constraints:
            raise ValueError(
                f"Algorithm {problem.algorithm.name} does not support "
                "nonlinear constraints."
            )

    # the following constraints will be handled via reparametrization
    constraints = [c for c in constraints if c["type"] != "nonlinear"]

    # ==================================================================================
    # Do first evaluation of user provided functions
    # ==================================================================================
    first_crit_eval = problem.fun_eval

    # do first derivative evaluation (if given)
    if problem.jac is not None:
        try:
            first_deriv_eval = problem.jac(problem.params)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as e:
            msg = "Error while evaluating derivative at start params."
            raise InvalidFunctionError(msg) from e

    if problem.fun_and_jac is not None:
        try:
            first_crit_and_deriv_eval = problem.fun_and_jac(problem.params)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as e:
            msg = "Error while evaluating criterion_and_derivative at start params."
            raise InvalidFunctionError(msg) from e

    if problem.jac is not None:
        used_deriv = first_deriv_eval
    elif problem.fun_and_jac is not None:
        used_deriv = first_crit_and_deriv_eval[1]
    else:
        used_deriv = None

    # ==================================================================================
    # Get the converter (for tree flattening, constraints and scaling)
    # ==================================================================================
    converter, internal_params = get_converter(
        params=problem.params,
        constraints=constraints,
        bounds=problem.bounds,
        func_eval=first_crit_eval.value,
        solver_type=problem.algorithm.algo_info.solver_type,
        scaling=problem.scaling,
        derivative_eval=used_deriv,
        add_soft_bounds=problem.multistart is not None,
    )

    # ==================================================================================
    # initialize the log database
    # ==================================================================================
    logger: LogStore[Any, Any] | None

    if problem.logging:
        logger = LogStore.from_options(problem.logging)
        problem_data = ProblemInitialization(problem.direction, problem.params)
        logger.problem_store.insert(problem_data)
    else:
        logger = None

    # ==================================================================================
    # Do some things that require internal parameters or bounds
    # ==================================================================================

    if converter.has_transforming_constraints and problem.multistart is not None:
        raise NotImplementedError(
            "multistart optimizations are not yet compatible with transforming "
            "constraints."
        )

    # get error penalty function
    error_penalty_func = get_error_penalty_function(
        start_x=internal_params.values,
        start_criterion=first_crit_eval,
        error_penalty=problem.error_penalty,
        solver_type=problem.algorithm.algo_info.solver_type,
        direction=problem.direction,
    )

    # process nonlinear constraints:
    internal_nonlinear_constraints = process_nonlinear_constraints(
        nonlinear_constraints=nonlinear_constraints,
        params=problem.params,
        bounds=problem.bounds,
        converter=converter,
        numdiff_options=problem.numdiff_options,
        skip_checks=problem.skip_checks,
    )

    x = internal_params.values
    internal_bounds = InternalBounds(
        lower=internal_params.lower_bounds,
        upper=internal_params.upper_bounds,
    )
    # ==================================================================================
    # Create a batch evaluator
    # ==================================================================================
    # TODO: Make batch evaluator an argument of maximize and minimize and move this
    # to create_optimization_problem
    batch_evaluator = process_batch_evaluator("joblib")

    # ==================================================================================
    # Create the InternalOptimizationProblem
    # ==================================================================================

    internal_problem = InternalOptimizationProblem(
        fun=problem.fun,
        jac=problem.jac,
        fun_and_jac=problem.fun_and_jac,
        converter=converter,
        solver_type=problem.algorithm.algo_info.solver_type,
        direction=problem.direction,
        bounds=internal_bounds,
        numdiff_options=problem.numdiff_options,
        error_handling=problem.error_handling,
        error_penalty_func=error_penalty_func,
        batch_evaluator=batch_evaluator,
        # TODO: Actually pass through linear constraints if possible
        linear_constraints=None,
        nonlinear_constraints=internal_nonlinear_constraints,
        logger=logger,
    )

    # ==================================================================================
    # Do actual optimization
    # ==================================================================================
    if problem.multistart is None:
        steps = [{"type": "optimization", "name": "optimization"}]

        # TODO: Actually use the step ids
        step_id = log_scheduled_steps_and_get_ids(  # noqa: F841
            steps=steps,
            logger=logger,
        )[0]

        raw_res = problem.algorithm.solve_internal_problem(internal_problem, x, step_id)

    else:
        multistart_options = get_internal_multistart_options_from_public(
            options=problem.multistart,
            params=problem.params,
            params_to_internal=converter.params_to_internal,
        )

        sampling_bounds = InternalBounds(
            lower=internal_params.soft_lower_bounds,
            upper=internal_params.soft_upper_bounds,
        )

        raw_res = run_multistart_optimization(
            local_algorithm=problem.algorithm,
            internal_problem=internal_problem,
            x=x,
            sampling_bounds=sampling_bounds,
            options=multistart_options,
            logger=logger,
            error_handling=problem.error_handling,
        )

    # ==================================================================================
    # Process the result
    # ==================================================================================

    _scalar_start_criterion = cast(
        float, first_crit_eval.internal_value(AggregationLevel.SCALAR)
    )
    log_reader: LogReader[Any] | None

    extra_fields = ExtraResultFields(
        start_fun=_scalar_start_criterion,
        start_params=problem.params,
        algorithm=problem.algorithm.algo_info.name,
        direction=problem.direction,
        n_free=internal_params.free_mask.sum(),
    )

    if problem.multistart is None:
        res = process_single_result(
            raw_res=raw_res,
            converter=converter,
            solver_type=problem.algorithm.algo_info.solver_type,
            extra_fields=extra_fields,
        )
    else:
        res = process_multistart_result(
            raw_res=raw_res,
            converter=converter,
            solver_type=problem.algorithm.algo_info.solver_type,
            extra_fields=extra_fields,
        )

    if logger is not None:
        assert problem.logging is not None
        log_reader = LogReader.from_options(problem.logging)
    else:
        log_reader = None

    res.logger = log_reader

    return res



"""Implement `nlopt` algorithms.

The documentation is heavily based on (nlopt documentation)[nlopt.readthedocs.io].

"""

from dataclasses import dataclass

import numpy as np
from numpy.typing import NDArray

from optimagic import mark
from optimagic.config import IS_NLOPT_INSTALLED
from optimagic.optimization.algo_options import (
    CONVERGENCE_FTOL_ABS,
    CONVERGENCE_FTOL_REL,
    CONVERGENCE_XTOL_ABS,
    CONVERGENCE_XTOL_REL,
    STOPPING_MAXFUN,
    STOPPING_MAXFUN_GLOBAL,
)
from optimagic.optimization.algorithm import Algorithm, InternalOptimizeResult
from optimagic.optimization.internal_optimization_problem import (
    InternalOptimizationProblem,
)
from optimagic.parameters.nonlinear_constraints import (
    equality_as_inequality_constraints,
)
from optimagic.typing import (
    AggregationLevel,
    NonNegativeFloat,
    PositiveInt,
)

if IS_NLOPT_INSTALLED:
    import nlopt


@mark.minimizer(
    name="nlopt_bobyqa",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_NLOPT_INSTALLED,
    is_global=False,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class NloptBOBYQA(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_XTOL_ABS
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_FTOL_ABS
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        res = _minimize_nlopt(
            problem=problem,
            x0=x0,
            is_global=self.algo_info.is_global,
            convergence_xtol_rel=self.convergence_xtol_rel,
            convergence_xtol_abs=self.convergence_xtol_abs,
            convergence_ftol_rel=self.convergence_ftol_rel,
            convergence_ftol_abs=self.convergence_ftol_abs,
            stopping_max_eval=self.stopping_maxfun,
            algorithm=nlopt.LN_BOBYQA,
        )

        return res


@mark.minimizer(
    name="nlopt_neldermead",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_NLOPT_INSTALLED,
    is_global=False,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class NloptNelderMead(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_XTOL_ABS
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_FTOL_ABS
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        res = _minimize_nlopt(
            problem=problem,
            x0=x0,
            is_global=self.algo_info.is_global,
            convergence_xtol_rel=self.convergence_xtol_rel,
            convergence_xtol_abs=self.convergence_xtol_abs,
            convergence_ftol_abs=self.convergence_ftol_abs,
            stopping_max_eval=self.stopping_maxfun,
            algorithm=nlopt.LN_NELDERMEAD,
        )

        return res


@mark.minimizer(
    name="nlopt_praxis",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_NLOPT_INSTALLED,
    is_global=False,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=False,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class NloptPRAXIS(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_XTOL_ABS
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_FTOL_ABS
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        res = _minimize_nlopt(
            problem=problem,
            x0=x0,
            is_global=self.algo_info.is_global,
            convergence_xtol_rel=self.convergence_xtol_rel,
            convergence_xtol_abs=self.convergence_xtol_abs,
            convergence_ftol_rel=self.convergence_ftol_rel,
            convergence_ftol_abs=self.convergence_ftol_abs,
            stopping_max_eval=self.stopping_maxfun,
            algorithm=nlopt.LN_PRAXIS,
        )

        return res


@mark.minimizer(
    name="nlopt_cobyla",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_NLOPT_INSTALLED,
    is_global=False,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=True,
    disable_history=False,
)
@dataclass(frozen=True)
class NloptCOBYLA(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_XTOL_ABS
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_FTOL_ABS
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        res = _minimize_nlopt(
            problem=problem,
            x0=x0,
            is_global=self.algo_info.is_global,
            convergence_xtol_rel=self.convergence_xtol_rel,
            convergence_xtol_abs=self.convergence_xtol_abs,
            convergence_ftol_rel=self.convergence_ftol_rel,
            convergence_ftol_abs=self.convergence_ftol_abs,
            stopping_max_eval=self.stopping_maxfun,
            algorithm=nlopt.LN_COBYLA,
        )

        return res


@mark.minimizer(
    name="nlopt_sbplx",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_NLOPT_INSTALLED,
    is_global=False,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class NloptSbplx(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_XTOL_ABS
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_FTOL_ABS
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        res = _minimize_nlopt(
            problem=problem,
            x0=x0,
            is_global=self.algo_info.is_global,
            convergence_xtol_rel=self.convergence_xtol_rel,
            convergence_xtol_abs=self.convergence_xtol_abs,
            convergence_ftol_rel=self.convergence_ftol_rel,
            convergence_ftol_abs=self.convergence_ftol_abs,
            stopping_max_eval=self.stopping_maxfun,
            algorithm=nlopt.LN_SBPLX,
        )

        return res


@mark.minimizer(
    name="nlopt_newuoa",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_NLOPT_INSTALLED,
    is_global=False,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class NloptNEWUOA(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_XTOL_ABS
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_FTOL_ABS
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        if problem.bounds.lower is None or problem.bounds.upper is None:
            algo = nlopt.LN_NEWUOA
        elif np.any(np.isfinite(problem.bounds.lower)) or np.any(
            np.isfinite(problem.bounds.upper)
        ):
            algo = nlopt.LN_NEWUOA_BOUND
        else:
            algo = nlopt.LN_NEWUOA
        res = _minimize_nlopt(
            problem=problem,
            x0=x0,
            is_global=self.algo_info.is_global,
            convergence_xtol_rel=self.convergence_xtol_rel,
            convergence_xtol_abs=self.convergence_xtol_abs,
            convergence_ftol_rel=self.convergence_ftol_rel,
            convergence_ftol_abs=self.convergence_ftol_abs,
            stopping_max_eval=self.stopping_maxfun,
            algorithm=algo,
        )

        return res


@mark.minimizer(
    name="nlopt_tnewton",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_NLOPT_INSTALLED,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class NloptTNewton(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_XTOL_ABS
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_FTOL_ABS
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        res = _minimize_nlopt(
            problem=problem,
            x0=x0,
            is_global=self.algo_info.is_global,
            convergence_xtol_rel=self.convergence_xtol_rel,
            convergence_xtol_abs=self.convergence_xtol_abs,
            convergence_ftol_rel=self.convergence_ftol_rel,
            convergence_ftol_abs=self.convergence_ftol_abs,
            stopping_max_eval=self.stopping_maxfun,
            algorithm=nlopt.LD_TNEWTON,
        )

        return res


@mark.minimizer(
    name="nlopt_lbfgsb",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_NLOPT_INSTALLED,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class NloptLBFGSB(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_XTOL_ABS
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_FTOL_ABS
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        res = _minimize_nlopt(
            problem=problem,
            x0=x0,
            is_global=self.algo_info.is_global,
            convergence_xtol_rel=self.convergence_xtol_rel,
            convergence_xtol_abs=self.convergence_xtol_abs,
            convergence_ftol_rel=self.convergence_ftol_rel,
            convergence_ftol_abs=self.convergence_ftol_abs,
            stopping_max_eval=self.stopping_maxfun,
            algorithm=nlopt.LD_LBFGS,
        )

        return res


@mark.minimizer(
    name="nlopt_ccsaq",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_NLOPT_INSTALLED,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class NloptCCSAQ(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_XTOL_ABS
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_FTOL_ABS
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        res = _minimize_nlopt(
            problem=problem,
            x0=x0,
            is_global=self.algo_info.is_global,
            convergence_xtol_rel=self.convergence_xtol_rel,
            convergence_xtol_abs=self.convergence_xtol_abs,
            convergence_ftol_rel=self.convergence_ftol_rel,
            convergence_ftol_abs=self.convergence_ftol_abs,
            stopping_max_eval=self.stopping_maxfun,
            algorithm=nlopt.LD_CCSAQ,
        )

        return res


@mark.minimizer(
    name="nlopt_mma",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_NLOPT_INSTALLED,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=True,
    disable_history=False,
)
@dataclass(frozen=True)
class NloptMMA(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_XTOL_ABS
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_FTOL_ABS
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        nonlinear_constraints = equality_as_inequality_constraints(
            problem.nonlinear_constraints
        )

        res = _minimize_nlopt(
            problem=problem,
            x0=x0,
            is_global=self.algo_info.is_global,
            convergence_xtol_rel=self.convergence_xtol_rel,
            convergence_xtol_abs=self.convergence_xtol_abs,
            convergence_ftol_rel=self.convergence_ftol_rel,
            convergence_ftol_abs=self.convergence_ftol_abs,
            stopping_max_eval=self.stopping_maxfun,
            algorithm=nlopt.LD_MMA,
            nonlinear_constraints=nonlinear_constraints,
        )

        return res


@mark.minimizer(
    name="nlopt_var",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_NLOPT_INSTALLED,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class NloptVAR(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_XTOL_ABS
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_FTOL_ABS
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN
    rank_1_update: bool = True

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        if self.rank_1_update:
            algo = nlopt.LD_VAR1
        else:
            algo = nlopt.LD_VAR2
        res = _minimize_nlopt(
            problem=problem,
            x0=x0,
            is_global=self.algo_info.is_global,
            convergence_xtol_rel=self.convergence_xtol_rel,
            convergence_xtol_abs=self.convergence_xtol_abs,
            convergence_ftol_rel=self.convergence_ftol_rel,
            convergence_ftol_abs=self.convergence_ftol_abs,
            stopping_max_eval=self.stopping_maxfun,
            algorithm=algo,
        )

        return res


@mark.minimizer(
    name="nlopt_slsqp",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_NLOPT_INSTALLED,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=True,
    disable_history=False,
)
@dataclass(frozen=True)
class NloptSLSQP(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_XTOL_ABS
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_FTOL_ABS
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        res = _minimize_nlopt(
            problem=problem,
            x0=x0,
            is_global=self.algo_info.is_global,
            convergence_xtol_rel=self.convergence_xtol_rel,
            convergence_xtol_abs=self.convergence_xtol_abs,
            convergence_ftol_rel=self.convergence_ftol_rel,
            convergence_ftol_abs=self.convergence_ftol_abs,
            stopping_max_eval=self.stopping_maxfun,
            algorithm=nlopt.LD_SLSQP,
        )

        return res


@mark.minimizer(
    name="nlopt_direct",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_NLOPT_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class NloptDirect(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_XTOL_ABS
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_FTOL_ABS
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN_GLOBAL
    locally_biased: bool = False
    random_search: bool = False
    unscaled_bounds: bool = False

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        if (
            not self.locally_biased
            and not self.random_search
            and not self.unscaled_bounds
        ):
            algo = nlopt.GN_DIRECT
        elif (
            self.locally_biased and not self.random_search and not self.unscaled_bounds
        ):
            algo = nlopt.GN_DIRECT_L
        elif self.locally_biased and not self.random_search and self.unscaled_bounds:
            algo = nlopt.GN_DIRECT_L_NOSCAL
        elif self.locally_biased and self.random_search and not self.unscaled_bounds:
            algo = nlopt.GN_DIRECT_L_RAND
        elif self.locally_biased and self.random_search and self.unscaled_bounds:
            algo = nlopt.GN_DIRECT_L_RAND_NOSCAL
        elif (
            not self.locally_biased and not self.random_search and self.unscaled_bounds
        ):
            algo = nlopt.GN_DIRECT_NOSCAL
        res = _minimize_nlopt(
            problem=problem,
            x0=x0,
            is_global=self.algo_info.is_global,
            convergence_xtol_rel=self.convergence_xtol_rel,
            convergence_xtol_abs=self.convergence_xtol_abs,
            convergence_ftol_rel=self.convergence_ftol_rel,
            convergence_ftol_abs=self.convergence_ftol_abs,
            stopping_max_eval=self.stopping_maxfun,
            algorithm=algo,
        )
        return res


@mark.minimizer(
    name="nlopt_esch",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_NLOPT_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class NloptESCH(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_XTOL_ABS
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_FTOL_ABS
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN_GLOBAL

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        res = _minimize_nlopt(
            problem=problem,
            x0=x0,
            is_global=self.algo_info.is_global,
            convergence_xtol_rel=self.convergence_xtol_rel,
            convergence_xtol_abs=self.convergence_xtol_abs,
            convergence_ftol_rel=self.convergence_ftol_rel,
            convergence_ftol_abs=self.convergence_ftol_abs,
            stopping_max_eval=self.stopping_maxfun,
            algorithm=nlopt.GN_ESCH,
        )

        return res


@mark.minimizer(
    name="nlopt_isres",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_NLOPT_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=True,
    disable_history=False,
)
@dataclass(frozen=True)
class NloptISRES(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_XTOL_ABS
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_FTOL_ABS
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN_GLOBAL

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        res = _minimize_nlopt(
            problem=problem,
            x0=x0,
            is_global=self.algo_info.is_global,
            convergence_xtol_rel=self.convergence_xtol_rel,
            convergence_xtol_abs=self.convergence_xtol_abs,
            convergence_ftol_rel=self.convergence_ftol_rel,
            convergence_ftol_abs=self.convergence_ftol_abs,
            stopping_max_eval=self.stopping_maxfun,
            algorithm=nlopt.GN_ISRES,
        )

        return res


@mark.minimizer(
    name="nlopt_crs2_lm",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_NLOPT_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class NloptCRS2LM(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_XTOL_ABS
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_FTOL_ABS
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN_GLOBAL
    population_size: PositiveInt | None = None

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        if self.population_size is None:
            population_size = 10 * (len(x0) + 1)
        else:
            population_size = self.population_size
        res = _minimize_nlopt(
            problem=problem,
            x0=x0,
            is_global=self.algo_info.is_global,
            convergence_xtol_rel=self.convergence_xtol_rel,
            convergence_xtol_abs=self.convergence_xtol_abs,
            convergence_ftol_rel=self.convergence_ftol_rel,
            convergence_ftol_abs=self.convergence_ftol_abs,
            stopping_max_eval=self.stopping_maxfun,
            algorithm=nlopt.GN_CRS2_LM,
            population_size=population_size,
        )

        return res


def _minimize_nlopt(
    problem,
    x0,
    algorithm,
    is_global,
    *,
    convergence_xtol_rel=None,
    convergence_xtol_abs=None,
    convergence_ftol_rel=None,
    convergence_ftol_abs=None,
    stopping_max_eval=None,
    population_size=None,
    nonlinear_constraints=None,
):
    """Run actual nlopt optimization argument, set relevant attributes."""

    def func(x, grad):
        if grad.size > 0:
            fun, jac = problem.fun_and_jac(x)
            grad[:] = jac
        else:
            fun = problem.fun(x)
        return fun

    if nonlinear_constraints is None:
        nonlinear_constraints = problem.nonlinear_constraints
    opt = nlopt.opt(algorithm, x0.shape[0])
    if convergence_ftol_rel is not None:
        opt.set_ftol_rel(convergence_ftol_rel)
    if convergence_ftol_abs is not None:
        opt.set_ftol_abs(convergence_ftol_abs)
    if convergence_xtol_rel is not None:
        opt.set_xtol_rel(convergence_xtol_rel)
    if convergence_xtol_abs is not None:
        opt.set_xtol_abs(convergence_xtol_abs)
    if problem.bounds.lower is not None:
        opt.set_lower_bounds(problem.bounds.lower)
    if problem.bounds.upper is not None:
        opt.set_upper_bounds(problem.bounds.upper)
    if stopping_max_eval is not None:
        opt.set_maxeval(stopping_max_eval)
    if population_size is not None:
        opt.set_population(population_size)
    if nonlinear_constraints:
        for constr in _get_nlopt_constraints(nonlinear_constraints, filter_type="eq"):
            opt.add_equality_mconstraint(constr["fun"], constr["tol"])
        for constr in _get_nlopt_constraints(nonlinear_constraints, filter_type="ineq"):
            opt.add_inequality_mconstraint(constr["fun"], constr["tol"])
    opt.set_min_objective(func)
    solution_x = opt.optimize(x0)
    return _process_nlopt_results(opt, solution_x, is_global)


def _process_nlopt_results(nlopt_obj, solution_x, is_global):
    messages = {
        1: "Convergence achieved ",
        2: (
            "Optimizer stopped because maximum value of criterion function was reached"
        ),
        3: (
            "Optimizer stopped because convergence_ftol_rel or "
            "convergence_ftol_abs was reached"
        ),
        4: (
            "Optimizer stopped because convergence_xtol_rel or "
            "convergence_xtol_abs was reached"
        ),
        5: "Optimizer stopped because max_criterion_evaluations was reached",
        6: "Optimizer stopped because max running time was reached",
        -1: "Optimizer failed",
        -2: "Invalid arguments were passed",
        -3: "Memory error",
        -4: "Halted because roundoff errors limited progress",
        -5: "Halted because of user specified forced stop",
    }
    success = nlopt_obj.last_optimize_result() in [1, 2, 3, 4]
    if is_global and not success:
        success = None
    processed = InternalOptimizeResult(
        x=solution_x,
        fun=nlopt_obj.last_optimum_value(),
        n_fun_evals=nlopt_obj.get_numevals(),
        success=success,
        message=messages[nlopt_obj.last_optimize_result()],
    )

    return processed


def _get_nlopt_constraints(constraints, filter_type):
    """Transform internal nonlinear constraints to NLOPT readable format."""
    filtered = [c for c in constraints if c["type"] == filter_type]
    nlopt_constraints = [_internal_to_nlopt_constaint(c) for c in filtered]
    return nlopt_constraints


def _internal_to_nlopt_constaint(c):
    """Sign flip description:

    In optimagic, inequality constraints are internally defined as g(x) >= 0. NLOPT uses
    h(x) <= 0, which is why we need to flip the sign.

    """
    tol = c["tol"]
    if np.isscalar(tol):
        tol = np.tile(tol, c["n_constr"])

    def _constraint(result, x, grad):
        result[:] = -c["fun"](x)  # see docstring for sign flip
        if grad.size > 0:
            grad[:] = -c["jac"](x)  # see docstring for sign flip

    new_constr = {
        "fun": _constraint,
        "tol": tol,
    }
    return new_constr
